@@ -1,53 +1,52 @@
+import datetime
 import json
 from copy import copy
-from enum import Enum
 from hashlib import sha256
 from json import JSONDecodeError
 from pathlib import Path
-from typing import List, Dict, Any, Optional, Union, NewType
-
-from .item_hash import ItemHash, ItemType
-
-try:
-    from typing import Literal
-except ImportError:
-    from typing_extensions import Literal
+from typing import Any, Dict, List, Literal, Optional, Type, TypeVar, Union, cast
 
 from pydantic import BaseModel, Extra, Field, validator
+from typing_extensions import TypeAlias
 
-from .abstract import BaseContent
-from .program import ProgramContent
+from .abstract import BaseContent, HashableModel
+from .base import Chain, HashType, MessageType
+from .execution.base import MachineType, Payment, PaymentType
+from .execution.instance import InstanceContent
+from .execution.program import ProgramContent
+from .item_hash import ItemHash, ItemType
 
-
-class Chain(str, Enum):
-    """Supported chains"""
-
-    AVAX = "AVAX"
-    BSC = "BSC"
-    CSDK = "CSDK"
-    DOT = "DOT"
-    ETH = "ETH"
-    NEO = "NEO"
-    NULS = "NULS"
-    NULS2 = "NULS2"
-    SOL = "SOL"
-    TEZOS = "TEZOS"
-
-
-class HashType(str, Enum):
-    """Supported hash functions"""
-
-    sha256 = "sha256"
-
-
-class MessageType(str, Enum):
-    """Message types supported by Aleph"""
-
-    post = "POST"
-    aggregate = "AGGREGATE"
-    store = "STORE"
-    program = "PROGRAM"
-    forget = "FORGET"
+__all__ = [
+    "AggregateContent",
+    "AggregateMessage",
+    "AlephMessage",
+    "AlephMessageType",
+    "Chain",
+    "ChainRef",
+    "ExecutableContent",
+    "ExecutableMessage",
+    "ForgetContent",
+    "ForgetMessage",
+    "HashType",
+    "InstanceContent",
+    "InstanceMessage",
+    "ItemHash",
+    "ItemType",
+    "HashableModel",
+    "MachineType",
+    "MessageConfirmation",
+    "MessageConfirmationHash",
+    "MessageType",
+    "Payment",
+    "PaymentType",
+    "PostContent",
+    "PostMessage",
+    "BaseContent",
+    "ProgramContent",
+    "ProgramMessage",
+    "StoreContent",
+    "StoreMessage",
+]
 
 
 class MongodbId(BaseModel):
@@ -70,7 +69,7 @@ class ChainRef(BaseModel):
     sender: str
     signature: str
     time: float
-    type = "POST"
+    type: Literal["POST"] = "POST"
 
 
 class MessageConfirmationHash(BaseModel):
@@ -90,7 +89,9 @@ class MessageConfirmation(BaseModel):
     # These two optional fields are introduced in recent versions of CCNs. They should
     # remain optional until the corresponding CCN upload (0.4.0) is widely uploaded.
     time: Optional[float] = None
-    publisher: Optional[str] = Field(default=None, description="The address that published the transaction.")
+    publisher: Optional[str] = Field(
+        default=None, description="The address that published the transaction."
+    )
 
     class Config:
         extra = Extra.forbid
@@ -147,6 +148,7 @@ class StoreContent(BaseContent):
     size: Optional[int] = None  # Generated by the node on storage
     content_type: Optional[str] = None  # Generated by the node on storage
     ref: Optional[str] = None
+    metadata: Optional[Dict[str, Any]] = Field(description="Metadata of the VM")
 
     class Config:
         extra = Extra.allow
@@ -196,7 +198,9 @@ class BaseMessage(BaseModel):
     size: Optional[int] = Field(
         default=None, description="Size of the content"
     )  # Almost always present
-    time: float = Field(description="Unix timestamp when the message was published")
+    time: datetime.datetime = Field(
+        description="Unix timestamp or datetime when the message was published"
+    )
     item_type: ItemType = Field(description="Storage method used for the content")
     item_content: Optional[str] = Field(
         default=None,
@@ -211,9 +215,11 @@ class BaseMessage(BaseModel):
     forgotten_by: Optional[List[str]]
 
     @validator("item_content")
-    def check_item_content(cls, v: Optional[str], values):
+    def check_item_content(cls, v: Optional[str], values) -> Optional[str]:
         item_type = values["item_type"]
-        if item_type == ItemType.inline:
+        if v is None:
+            return None
+        elif item_type == ItemType.inline:
             try:
                 json.loads(v)
             except JSONDecodeError:
@@ -221,14 +227,13 @@ class BaseMessage(BaseModel):
                     "Field 'item_content' does not appear to be valid JSON"
                 )
         else:
-            if v is not None:
-                raise ValueError(
-                    f"Field 'item_content' cannot be defined when 'item_type' == '{item_type}'"
-                )
+            raise ValueError(
+                f"Field 'item_content' cannot be defined when 'item_type' == '{item_type}'"
+            )
         return v
 
     @validator("item_hash")
-    def check_item_hash(cls, v, values):
+    def check_item_hash(cls, v: ItemHash, values) -> ItemHash:
         item_type = values["item_type"]
         if item_type == ItemType.inline:
             item_content: str = values["item_content"]
@@ -257,6 +262,13 @@ class BaseMessage(BaseModel):
             raise ValueError("Message cannot be 'confirmed' without 'confirmations'")
         return v
 
+    @validator("time")
+    def convert_float_to_datetime(cls, v, values):
+        if isinstance(v, float):
+            v = datetime.datetime.fromtimestamp(v)
+        assert isinstance(v, datetime.datetime)
+        return v
+
     class Config:
         extra = Extra.forbid
         exclude = {"id_", "_id"}
@@ -312,25 +324,46 @@ class ProgramMessage(BaseMessage):
         return v
 
 
-message_types = (
+class InstanceMessage(BaseMessage):
+    type: Literal[MessageType.instance]
+    content: InstanceContent
+
+
+AlephMessage: TypeAlias = Union[
     PostMessage,
     AggregateMessage,
     StoreMessage,
     ProgramMessage,
+    InstanceMessage,
     ForgetMessage,
-)
-
-AlephMessage = NewType("AlephMessage", Union[message_types])
+]
 
 
-def Message(**message_dict: Dict) -> AlephMessage:
+T = TypeVar("T", bound=AlephMessage)
+
+AlephMessageType: TypeAlias = Type[T]
+
+message_classes: List[AlephMessageType] = [
+    PostMessage,
+    AggregateMessage,
+    StoreMessage,
+    ProgramMessage,
+    InstanceMessage,
+    ForgetMessage,
+]
+
+ExecutableContent: TypeAlias = Union[InstanceContent, ProgramContent]
+ExecutableMessage: TypeAlias = Union[InstanceMessage, ProgramMessage]
+
+
+def parse_message(message_dict: Dict) -> AlephMessage:
     """Returns the message class corresponding to the type of message."""
-    for message_class in message_types:
+    for message_class in message_classes:
         message_type: MessageType = MessageType(
             message_class.__annotations__["type"].__args__[0]
         )
         if message_dict["type"] == message_type:
-            return message_class(**message_dict)
+            return message_class.parse_obj(message_dict)
     else:
         raise ValueError(f"Unknown message type {message_dict['type']}")
 
@@ -349,27 +382,36 @@ def add_item_content_and_hash(message_dict: Dict, inplace: bool = False):
 
 
 def create_new_message(
-    message_dict: Dict, factory: Union[Message, AlephMessage] = Message
-) -> AlephMessage:
+    message_dict: Dict,
+    factory: Optional[Type[T]] = None,
+) -> T:
     """Create a new message from a dict.
     Computes the 'item_content' and 'item_hash' fields.
     """
-    return factory(**add_item_content_and_hash(message_dict))
+    message_content = add_item_content_and_hash(message_dict)
+    if factory:
+        return cast(T, factory.parse_obj(message_content))
+    else:
+        return cast(T, parse_message(message_content))
 
 
 def create_message_from_json(
-    json_data: str, factory: Union[Message, AlephMessage] = Message
+    json_data: str,
+    factory: Optional[AlephMessageType] = None,
 ) -> AlephMessage:
     """Create a new message from a JSON encoded string.
     Computes the 'item_content' and 'item_hash' fields.
     """
     message_dict = json.loads(json_data)
-    add_item_content_and_hash(message_dict, inplace=True)
-    return factory(**message_dict)
+    message_content = add_item_content_and_hash(message_dict, inplace=True)
+    if factory:
+        return factory.parse_obj(message_content)
+    else:
+        return parse_message(message_content)
 
 
 def create_message_from_file(
-    filepath: Path, factory: Union[Message, AlephMessage] = Message, decoder=json
+    filepath: Path, factory: Optional[AlephMessageType] = None, decoder=json
 ) -> AlephMessage:
     """Create a new message from an encoded file.
     Expects json by default, but allows other decoders with a method `.load()`
@@ -378,8 +420,11 @@ def create_message_from_file(
     """
     with open(filepath) as fd:
         message_dict = decoder.load(fd)
-    add_item_content_and_hash(message_dict, inplace=True)
-    return factory(**message_dict)
+    message_content = add_item_content_and_hash(message_dict, inplace=True)
+    if factory:
+        return factory.parse_obj(message_content)
+    else:
+        return parse_message(message_content)
 
 
 class MessagesResponse(BaseModel):
